
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_db.tools.db.query.base.query_builder import QueryBuilder, DBCompareOperator
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing


logger = logging.getLogger(__name__)

try:
    import pypika.functions
    from pypika.queries import Table, Query
    from pypika.terms import Parameter, Function, Index

    with_pypika = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"PyPika is not available for QueryBuilder. Initialization failed on error: {exc}")
    with_pypika = False



class PypikaQueryBuilder(QueryBuilder):
    """
    Query builder for PyPika library.
    """
    @classmethod
    def is_available(cls):
        return with_pypika
    
    def __init__(self, name, db_client=None):
        super().__init__(name, db_client)
    
    def select(self, table_name, where_data: dict=None, where_compare_data: list=None, sql_return="*"):
        """
        Simple query & values builder of a select by filtering on given where data.
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        table = Table(table_name)
        res = Query.from_(table)
        values = ()
        
        if sql_return == "count(*)":
            res = res.select(pypika.functions.Count("*"))
        elif isinstance(sql_return, list):
            res = res.select(*sql_return)
        else:
            res = res.select(sql_return)
        
        if where_data:
            res, values = self.where(res, values, where_data)
        if where_compare_data:
            res, values = self.where_compare(res, values, list_of_field_operator_value=where_compare_data)
        
        return res, values
    
    def insert(self, table_name, data: dict):
        """
        Simple query & values builder of an insert of given data.
        @param data: data to insert as dictionary of (field_name, value).
        """
        col_names = tuple(sorted(data.keys()))
        values = tuple((data[c] for c in col_names))
        sql_placeholder = self.db_client._get_sql_placeholder()
        
        values_placeholder = [Parameter(sql_placeholder)] * len(col_names)
        res = Query.into(table_name).columns(*col_names).insert(*values_placeholder)
        
        return res, values
    
    def update(self, table_name, data: dict, where_data: dict=None, where_compare_data: list=None):
        """
        Simple query & values builder of an update of given data.
        @param data: data to update as dictionary of (field_name, value).
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        col_names = tuple(sorted(data.keys()))
        sql_placeholder = self.db_client._get_sql_placeholder()
        
        table = Table(table_name)
        res = Query.update(table)
        values = []
        for c in col_names:
            val = data[c]
            if isinstance(val, Function):
                res = res.set(c, val)
            else:
                res = res.set(c, Parameter(sql_placeholder))
                values.append(val)
            
        if where_data:
            res, values = self.where(res, values, where_data)
        if where_compare_data:
            res, values = self.where_compare(res, values, list_of_field_operator_value=where_compare_data)
        
        return res, values
    
    def delete(self, table_name, where_data: dict=None, where_compare_data: list=None):
        """
        Simple query & values builder of a delete by filtering on given where data.
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        table = Table(table_name)
        res = Query.from_(table).delete()
        values = ()
        
        if where_data:
            res, values = self.where(res, values, where_data)
        if where_compare_data:
            res, values = self.where_compare(res, values, list_of_field_operator_value=where_compare_data)
            
        return res, values
    
    def where(self, query, values, where_data: dict):
        col_names = tuple(sorted(where_data.keys()))
        sql_placeholder = self.db_client._get_sql_placeholder()
        
        res = query
        table = self.__get_table(query)
        where_values = []
        for c in col_names:
            val = where_data[c]
            if val is None:
                res, where_values = self.where_is_null(res, where_values, c)
            else:
                res = res.where(getattr(table, c) == Parameter(sql_placeholder))
                where_values.append(where_data[c])
        
        if values is not None:
            values = (*values, *where_values)
        else:
            values = where_values
        return res, values
    
    def where_compare(self, query, values, *, field_name=None, operator:DBCompareOperator=None, value=None, list_of_field_operator_value=None):
        if list_of_field_operator_value is not None:
            return super().where_compare(query, values, list_of_field_operator_value=list_of_field_operator_value)
        
        # Manage operators calling other methods
        if operator == DBCompareOperator.In:
            return self.where_in(query, values, field_name, value, not_in=False)
        elif operator == DBCompareOperator.NotIn:
            return self.where_in(query, values, field_name, value, not_in=True)
        
        # Manage other operators
        table = self.__get_table(query)
        sql_placeholder = self.db_client._get_sql_placeholder()
        
        if operator == DBCompareOperator.Different:
            res = query.where(getattr(table, field_name) != Parameter(sql_placeholder))
        elif operator == DBCompareOperator.Equal:
            res = query.where(getattr(table, field_name) == Parameter(sql_placeholder))
        elif operator == DBCompareOperator.Inferior:
            res = query.where(getattr(table, field_name) < Parameter(sql_placeholder))
        elif operator == DBCompareOperator.InferiorOrEqual:
            res = query.where(getattr(table, field_name) <= Parameter(sql_placeholder))
        elif operator == DBCompareOperator.Superior:
            res = query.where(getattr(table, field_name) > Parameter(sql_placeholder))
        elif operator == DBCompareOperator.SuperiorOrEqual:
            res = query.where(getattr(table, field_name) >= Parameter(sql_placeholder))
        else:
            raise TechnicalException(f"Unmanaged compare operator {operator}")
            
        if values is not None:
            values = (*values, value)
        else:
            values = (value)
        return res, values
    
    def where_in(self, query, values, field_name, field_values, not_in=False):
        table = self.__get_table(query)
        if not_in:
            res = query.where(getattr(table, field_name).notin(field_values))
        else:
            res = query.where(getattr(table, field_name).isin(field_values))
        return res, values
    
    def where_is_null(self, query, values, field_name, is_not_null=False):
        table = self.__get_table(query)
        if is_not_null:
            res = query.where(getattr(table, field_name).notnull())
        else:
            res = query.where(getattr(table, field_name).isnull())
        return res, values
        
    def where_json_value(self, query, values, field_name, key, value, as_text_value=False):
        table = self.__get_table(query)
        sql_placeholder = self.db_client._get_sql_placeholder()
        
        if as_text_value:
            res = query.where(getattr(table, field_name).get_text_value(key) == Parameter(sql_placeholder))
        else:
            res = query.where(getattr(table, field_name).get_json_value(key) == Parameter(sql_placeholder))
        
        if values is not None:
            values = (*values, value)
        else:
            values = (value)
        return res, values
    
    def to_sql(self, query):
        if query.__module__.startswith('pypika'):
            return query.get_sql()
        else:
            raise TechnicalException(f"Unmanaged query of type {Typing.get_object_class_fullname(query)}")

    def create_index(self, index_name, table_name, column_names, unique=False, if_not_exists=False):
        index = Index(index_name)
        table = Table(table_name)
        res = Query \
            .create_index(index) \
            .on(table) \
            .columns(*column_names)
        if unique:
            res = res.unique()
        if if_not_exists:
            res = res.if_not_exists()
        return res

    def drop_index(self, index_name, if_exists=False):
        index = Index(index_name)
        res = Query.drop_index(index)
        if if_exists:
            res = res.if_exists()
        return res

    # def is_query_type(self, query, query_type):
    #     query_type = query_type.lower()
    #
    #     if query_type == 'insert':
    #         return query._insert_table is not None
    #     elif query_type == 'select':
    #         return len(query._selects) > 0
    #     elif query_type == 'update':
    #         return query._update_table is not None
    #     elif query_type == 'delete':
    #         return query._delete_from
    #     else:
    #         raise TechnicalException(f"Unmanaged query type '{query_type}' (managed query types: 'insert', 'select', 'update', 'delete')")
        
    def __get_table(self, query):
        if query._from is not None and len(query._from) > 0:
            return query._from[0]
        elif query._insert_table is not None:
            return query._insert_table
        elif query._update_table is not None:
            return query._update_table
        else:
            raise TechnicalException(f"Failed to extract table from query [{query}]: {Tools.represent_object(query)}")



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
import string
from holado_core.common.exceptions.functional_exception import FunctionalException
import re
from holado_python.standard_library.typing import Typing

try:
    from bitarray import bitarray
    with_bitarray = True
except:
    with_bitarray = False

logger = logging.getLogger(__name__)


class StrTools(object):
    
    #TODO EKL: move following methods from Tools to this class
    # @classmethod
    # def get_indent_string(cls, indent):
    #     return " " * indent
    #
    # @classmethod
    # def indent_string(cls, indent, txt):
    #     ind_str = StrTools.get_indent_string(indent)
    #     lines = txt.split("\n") if txt else []
    #     return ind_str + ("\n" + ind_str).join(lines)
    #
    # @classmethod
    # def truncate_text(cls, text, length = Config.message_truncate_length, truncated_suffix = "[...]", is_length_with_suffix=False):
    #     if len(text) > length:
    #         if truncated_suffix:
    #             if is_length_with_suffix:
    #                 return text[0 : length - len(truncated_suffix)] + truncated_suffix
    #             else:
    #                 return text[0 : length] + truncated_suffix
    #         else:
    #             return text[0 : length]
    #     else:
    #         return text
    
    if with_bitarray:
        @classmethod
        def is_bitarray(cls, src):
            # Convert source to str
            if isinstance(src, bytes):
                value_str = src.decode('utf-8')
            elif isinstance(src, str):
                value_str = src
            else:
                raise FunctionalException(f"Unexpected source type {Typing.get_object_class_fullname(src)} (allowed types: string, bytes)")
            
            try:
                bitarray(value_str)
            except:
                return False
            return True
    
    @classmethod
    def is_hex(cls, src):
        if not isinstance(src, str):
            return False
        
        hex_digits = set(string.hexdigits)
        return all(c in hex_digits for c in src)
    
    @classmethod
    def hex_to_bytes(cls, src):
        if not cls.is_hex(src):
            raise FunctionalException(f"Source [{src}] (type: {Typing.get_object_class_fullname(src)}) is not an hexadecimal string")
        
        if len(src) % 2 != 0:
            raise FunctionalException(f"Hex string must have a length multiple of 2. If not, a padding is usually applied with a 0 at left or right. Hex string: '{src}'")
        
        try:
            res = bytes.fromhex(src)
        except ValueError as exc:
            raise TechnicalException(f"Error with hex string '{src}'") from exc
        
        # if len(res) * 2 != len(src):
        #     raise TechnicalException(f"Failed to convert hex to bytes: hex length is {len(src)} ; bytes length is {len(res)}")
        
        return res
    
    @classmethod
    def to_bytes(cls, src):
        if isinstance(src, int):
            signed = True if src < 0 else False
        
            length = 1
            val = src
            while val > 255:
                length += 1
                val = val / 256
        
            res = src.to_bytes(length, 'big', signed=signed)
        elif isinstance(src, bytes):
            res = src
        elif isinstance(src, str):
            if re.match(r"^b('|\").*\1$", src):
                res = eval(src)
                if not isinstance(res, bytes):
                    raise TechnicalException(f"Failed to eval string [{src}] as bytes ; eval result: [{res}] (type: {Typing.get_object_class_fullname(res)})")
            else:
                res = src.encode('utf-8')
        else:
            raise FunctionalException(f"Unexpected source type {Typing.get_object_class_fullname(src)} (allowed types: int, string, bytes)")
        
        return res
    
    @classmethod
    def to_hex(cls, src, do_upper=True):
        src_bytes = cls.to_bytes(src)
        
        res = src_bytes.hex()
        if do_upper:
            res = res.upper()
        
        return res
    
    @classmethod
    def to_string(cls, src):
        if isinstance(src, bytes):
            return src.decode('utf-8')
        elif isinstance(src, str):
            return src
        else:
            return str(src)





# -*- coding: utf-8 -*-

import logging
from holado.common.context.session_context import SessionContext
from holado_system.system.command.command import Command, CommandStates
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_test.scenario.step_tools import StepTools
from holado_core.tools.abstracts.blocking_command_service import BlockingCommandService
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)

def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()

@Given(r"(?P<var_name>{Variable}) = new command (?P<cmd>{AnyStr})(?<! with)")
def step_impl(context, var_name, cmd):
    var_name = StepTools.evaluate_variable_name(var_name)
    cmd = StepTools.evaluate_string_parameter(cmd)
    obj_cmd = Command(cmd)
    __get_variable_manager().register_variable(var_name, obj_cmd)

@Given(r"(?P<var_name>{Variable}) = new command (?P<cmd>{AnyStr}) with")
def step_impl(context, var_name, cmd):
    var_name = StepTools.evaluate_variable_name(var_name)
    cmd = StepTools.evaluate_scenario_parameter(cmd)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    arguments_dict = ValueTableConverter.convert_name_value_table_2_dict(table)
    
    
    blocking = False
    auto_stop = False
    do_log_output = False
    do_raise_on_stderr = False
    name = ''
    
    if 'blocking' in arguments_dict:
        blocking = arguments_dict['blocking']
        if 'name' in arguments_dict:
            name = arguments_dict['name']
        if 'auto_stop' in arguments_dict:
            auto_stop = arguments_dict['auto_stop']
    else:
        if 'name' in arguments_dict:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Argument 'name' is useless for non-blocking commands")
        if 'auto_stop' in arguments_dict:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Argument 'auto_stop' is useless for non-blocking commands")
    if 'do_log_output' in arguments_dict:
        do_log_output = arguments_dict['do_log_output']
    if 'do_raise_on_stderr' in arguments_dict:
        do_raise_on_stderr = arguments_dict['do_raise_on_stderr']
    
    if blocking:
        obj_cmd = BlockingCommandService(name, cmd, do_log_output=do_log_output, do_raise_on_stderr=do_raise_on_stderr)
        obj_cmd.auto_stop = auto_stop
    else:
        obj_cmd = Command(cmd, do_log_output=do_log_output, do_raise_on_stderr=do_raise_on_stderr)
    
    __get_variable_manager().register_variable(var_name, obj_cmd)

@Then(r"command (?P<var_name>{Variable}) has status (?P<expected_status>Ready|Running|Success|Error)")
def step_impl(context, var_name, expected_status):
    obj_cmd =  StepTools.evaluate_variable_value(var_name)
    
    if isinstance(obj_cmd, Command):
        state = obj_cmd.state
    elif isinstance(obj_cmd, BlockingCommandService):
        state = obj_cmd.status
    else:
        raise TechnicalException(f"Variable {var_name} is not a command")
        
    if state is not CommandStates[expected_status]:
        raise FunctionalException(f"Command state error : expected {expected_status} found {state.name}")
    
@Given(r"run command (?P<var_name>{Variable})")
def step_impl(context, var_name):
    obj_cmd =  StepTools.evaluate_variable_value(var_name)
    obj_cmd.start()
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
from holado.common.context.context import Context
import logging
from holado_scripting.common.tools.variable_manager import VariableManager
from holado.common.context.session_context import SessionContext

logger = logging.getLogger(__name__)


class ThreadContext(Context):
    def __init__(self, name, unique_name):
        super().__init__(f"Thread[{name};{unique_name}]")
        
        self.__thread_uid = None
        self.__thread_name = name
        self.__thread_unique_name = unique_name
    
    def __str__(self):
        return f"{{ThreadContext({id(self)}):{self.__thread_uid}:{self.__thread_name}:{self.__thread_unique_name}}}"

    @property
    def thread_name(self):
        return self.__thread_name
    
    @property
    def thread_unique_name(self):
        return self.__thread_unique_name
    
    @property
    def thread_uid(self):
        return self.__thread_uid
    
    @thread_uid.setter
    def thread_uid(self, uid):
        self.__thread_uid = uid
    
    
    def has_variable_manager(self):
        return self.has_object("variable_manager")
        
    def get_variable_manager(self) -> VariableManager:
        if not self.has_variable_manager():
            parent_thread_context = SessionContext.instance().multitask_manager.get_parent_thread_context(uid=self.thread_uid)
            if parent_thread_context is not None:
                var_man = parent_thread_context.get_variable_manager()
            else:
                var_man = SessionContext.instance().variable_manager
                
            manager = VariableManager(var_man)
            self.set_object("variable_manager", manager)
            file_path = SessionContext.instance().report_manager.get_path("logs", f"variable_update-thread_{self.thread_uid}.log") if SessionContext.instance().with_session_path else None
            manager.initialize(SessionContext.instance().dynamic_text_manager, SessionContext.instance().unique_value_manager,
                               variable_update_log_file_path=file_path)
        return self.get_object("variable_manager")
    
    
    ### Manage feature context
    
    def has_feature_context(self):
        return self.has_object("feature_context")
    
    def get_feature_context(self):
        return self.get_object("feature_context")
    
    def set_feature_context(self, feature_context):
        return self.set_object("feature_context", feature_context)
    
    def delete_feature_context(self):
        self.get_feature_context().delete_object()
        self.remove_object("feature_context")
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.handlers.redo import Redo
from holado.common.handlers.undefined import undefined_value
from holado.holado_config import Config

logger = logging.getLogger(__name__)


class WaitOnEventResult(Redo):
    def __init__(self, name, on_event_func):
        super().__init__(name)
        self.__on_event_func = on_event_func
        self.__on_event_func_result = None
        
    def _process(self):
        return self.__on_event_func_result
    
    def on_event(self, *args, **kwargs):
        self.__on_event_func_result = self.__on_event_func(*args, **kwargs)


class WaitIsTrueOnEvent(WaitOnEventResult):
    def __init__(self, name, on_event_func):
        super().__init__(name, on_event_func)
        self.redo_until(True)


class WaitFuncResult(Redo):
    def __init__(self, name, func):
        super().__init__(name)
        self.__func = func
        
    def _process(self):
        return self.__func()

class WaitFuncResultVerifying(WaitFuncResult):
    def __init__(self, name, result_func, result_verify_func):
        super().__init__(name, result_func)
        self.__result_verify_func = result_verify_func
        
    def _is_redo_needed(self, result):
        return not self.__result_verify_func(result)


class WaitChange(WaitFuncResult):
    def __init__(self, name, func, *, do_process_in_thread=True, result_reference=undefined_value, timeout_seconds=None, polling_seconds=None):
        super().__init__(name, func)
        self.with_process_in_thread(do_process_in_thread)
        self.__result_reference = result_reference
        
        # Compute default timeout and polling if needed
        if timeout_seconds is None:
            timeout_seconds = Config.timeout_seconds
        if polling_seconds is None:
            polling_seconds = timeout_seconds / 1000
            if polling_seconds > 0.01:
                polling_seconds = 0.01
                
        self.with_timeout(timeout_seconds)
        self.polling_every(polling_seconds)
        
    def _is_redo_needed(self, result):
        if self.__result_reference is undefined_value:
            # Update reference with first run result
            self.__result_reference = result
            return True
        else:
            return result == self.__result_reference

class WaitEndChange(WaitFuncResult):
    def __init__(self, name, func, *, do_process_in_thread=True, result_reference=undefined_value, first_timeout_seconds=None, window_seconds=None, polling_seconds=None):
        super().__init__(name, func)
        self.with_process_in_thread(do_process_in_thread)
        self.__result_reference = result_reference
        
        # Compute default timeouts and duration if needed
        if window_seconds is None:
            if first_timeout_seconds is not None and first_timeout_seconds > 0:
                window_seconds = first_timeout_seconds / 10
            else:
                window_seconds = 1
            if window_seconds < 0.01:
                window_seconds = 0.01       # a window period below 10 ms is usually not efficient in testing context
        if polling_seconds is None:
            min_window = min(first_timeout_seconds, window_seconds) if first_timeout_seconds is not None else window_seconds
            polling_seconds = min_window / 100       # 1% of error on window detection
            if polling_seconds > 0.1:
                polling_seconds = 0.1       # a polling period over 100 ms is usually not efficient in testing context
            elif polling_seconds < 0.001:
                polling_seconds = 0.001       # a polling period below 1 ms is usually not efficient in testing context
        if first_timeout_seconds is None:
            first_timeout_seconds = window_seconds
        
        # Start with a timeout equal to first_timeout_seconds
        self.with_timeout(first_timeout_seconds)
        
        self.polling_every(polling_seconds)
        self.__window_seconds = window_seconds
    
    def _execute_after_process(self, result):
        if self.__result_reference is undefined_value:
            # Update reference with first run result
            self.__result_reference = result
        elif result != self.__result_reference:
            self.__result_reference = result
            
            # As a change is received, update timeout with window duration and do not raise on timeout anymore
            self.update_timeout(self.__window_seconds)
            self.without_raise_on_timeout()
        
    def _is_redo_needed(self, result):
        return True






import logging
import os
from holado_core.common.tools.path_manager import PathManager

logger = logging.getLogger(__name__)


class TSPathManager(PathManager):

    def __init__(self):
        super().__init__()
        
    def initialize(self):
        pass
    
    def __get_root_path(self):
        here = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(here, "..", "..", "..")
        
    def get_reports_path(self, name=None):
        if os.getenv('TESTING_SOLUTION_REPORTS_PATH') is not None:
            reports_root = os.getenv('TESTING_SOLUTION_REPORTS_PATH')
        elif os.getenv('TEST_OUTPUT_BASEDIR') is not None:
            reports_root = os.path.join(os.getenv('TEST_OUTPUT_BASEDIR'), "reports")
        else:
            reports_root = super().get_reports_path()

        if name is not None:
            return os.path.join(reports_root, name)
        else:
            return reports_root

    def get_static_files_path(self, name=None):
        root_path = self.__get_root_path()
        res = os.path.normpath(os.path.join(root_path, "resources", "static_files"))
        if name is not None:
            res = os.path.join(res, name)
        return res
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################



from builtins import super
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
import logging
from holado_core.common.tables.comparators.string_table_row_comparator import StringTableRowComparator
from holado_core.common.exceptions.exceptions import Exceptions
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.comparators.table_comparator import TableComparator

logger = logging.getLogger(__name__)


class TableWithHeaderComparator(TableComparator):
    def __init__(self, header_comparator=None, row_comparator=None, cell_comparator=None):
        super().__init__(row_comparator=row_comparator, cell_comparator=cell_comparator)
        
        self.__header_comparator = header_comparator if header_comparator is not None else StringTableRowComparator()

    @property
    def header_comparator(self):
        return self.__header_comparator
        
    def equals(self, table_1, table_2, is_obtained_vs_expected = True, raise_exception = True):
        return (self.equals_headers(table_1, table_2, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
                and super().equals(table_1, table_2, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception))
    
    def contains_rows(self, table_1, table_2, check_row_order = True, is_obtained_vs_expected = True, raise_exception = True):
        return (self.equals_headers(table_1, table_2, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
                and super().contains_rows(table_1, table_2, check_row_order=check_row_order, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception))

    def doesnt_contain_rows(self, table_1, table_2, is_obtained_vs_expected = True, raise_exception = True):
        return (self.equals_headers(table_1, table_2, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception)
                and super().doesnt_contain_rows(table_1, table_2, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception))

    def equals_headers(self, table_1, table_2, is_obtained_vs_expected = True, raise_exception = True):
        try:
            return self.__header_comparator.equals(table_1.header, table_2.header, is_obtained_vs_expected=is_obtained_vs_expected, raise_exception=raise_exception);
        except FunctionalException as exc:
            msg = f"Table headers are not the same:\n -> {Exceptions.exception_message(exc)}"
            if raise_exception:
                raise VerifyException(msg) from exc
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(msg)
                return False
        except Exception as exc:
            msg = "Error while comparing tables headers:\n{}\n  table {} header: {}\n  table {} header: {}".format(
                    Tools.indent_string(4, Exceptions.exception_message(exc)), 
                    self._get_name_1(is_obtained_vs_expected), self._represent_row(table_1.header, 0), 
                    self._get_name_2(is_obtained_vs_expected), self._represent_row(table_2.header, 0))
            raise TechnicalException(msg) from exc
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.handlers.object import DeleteableObject
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.finders.tools.find_builder import FindBuilder
from holado.common.handlers.enums import ObjectStates

logger = logging.getLogger(__name__)



class Driver(DeleteableObject):
    """ Base class for drivers.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name=None):
        super().__init__(name or "Driver")
        
        self.__internal_driver = None
        self.__internal_api = None
        
        # Tools
        self.__text_verifier = None
        self.__find_builder = None      # Builder used to create default find context and parameters
    
    @property
    def _has_internal_driver(self):
        """
        @return If internal driver is defined
        """
        return self.__internal_driver is not None
    
    @property
    def internal_driver(self):
        """
        @return Internal driver
        """
        if not self._has_internal_driver():
            raise TechnicalException("Internal driver is not defined.")
        return self.__internal_driver
    
    @internal_driver.setter
    def internal_driver(self, internal_driver):
        self.__internal_driver = internal_driver
    
    @property
    def internal_api(self):
        """
        @return Internal API
        """
        if self.__internal_api is None:
            raise TechnicalException("Internal API is not defined.")
        return self.__internal_api
    
    @property
    def text_verifier(self):
        """
        @return Text verifier to use
        """
        if self.__text_verifier is None:
            raise TechnicalException("Text verifier is not defined")
        return self.__text_verifier
    
    @text_verifier.setter
    def text_verifier(self, text_verifier):
        """
        @param text_verifier Text verifier to use
        """
        self.__text_verifier = text_verifier
    
    @property
    def find_builder(self):
        """
        @return Find builder (if not set a default one is instantiated)
        """
        if self.__find_builder is None:
            self.__find_builder = FindBuilder()
        return self.__find_builder
    
    @find_builder.setter
    def find_builder(self, find_builder):
        """
        @param find_builder Find builder to use rather than default
        """
        self.__find_builder = find_builder
    
    def initialize(self):
        """
        Initialize driver
        """
        self.__internal_api = self._initialize_internal_api()
    
    def _initialize_internal_api(self):
        """
        Method used internally at driver initialization, in order to instantiate the appropriate internal API type.
        @return a new internal API instance
        """
        raise NotImplementedError

    def is_open(self):
        """
        @return True if driver is open
        """
        return self._has_internal_driver()
    
    def _verify_is_open(self):
        if not self.is_open():
            raise FunctionalException("Driver is not opened")
    
    def open(self):
        """
        Open driver
        """
        self.object_state = ObjectStates.Opening
        self._close_driver()
        self.object_state = ObjectStates.Open

    def _open_driver(self):
        """
        Implement open process
        """
        raise NotImplementedError
    
    def close(self):
        """
        Close driver
        """
        self.object_state = ObjectStates.Closing
        self._close_driver()
        self.object_state = ObjectStates.Closed
    
    def _close_driver(self):
        """
        Implement close process
        """
        raise NotImplementedError
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.handlers.object import DeleteableObject
from holado_multitask.multithreading.functionthreaded import FunctionThreaded
from holado_python.common.tools.datetime import DateTime
import json
from holado_json.filesystem.stream_json_file import StreamJSONFile
import re
from holado.common.context.session_context import SessionContext
import os
from datetime import datetime
import copy
from holado_system.system.filesystem.file import File
from holado.common.handlers.undefined import is_undefined, default_value
import time

logger = logging.getLogger(__name__)


def _get_timestamps_field_name(input_field_name):
    if input_field_name is None:
        return None
    elif is_undefined(input_field_name):
        return 'time'
    else:
        return input_field_name

def _get_include_others_name(input_include_others_name):
    if input_include_others_name is None:
        return None
    elif is_undefined(input_include_others_name):
        return '[others]'
    else:
        return input_include_others_name


class DockerContainerLogsFollower(DeleteableObject):
    """Generic follower of logs of a container
    """
    def __init__(self, docker_container):
        super.__init__(self, docker_container.container.name)
        self.__docker_container = docker_container
        self.__logs = []
        self.__logs_stream = None
        self.__thread = None
    
    def _delete_object(self):
        self.close()
    
    def close(self):
        if self.__logs_stream is not None:
            self.__logs_stream.close()
            self.__thread.join()
            
            self.__logs_stream = None
            self.__thread = None
    
    def follow_logs(self, stdout=True, stderr=True, tail='all', since=None):
        """Follow logs from Docker Engine
        """
        if isinstance(since, str):
            since = DateTime.str_2_datetime(since)
        
        if self.__logs_stream is not None:
            raise TechnicalException(f"Logs of container '{self.name}' are already followed")
        self.__logs_stream = self.__docker_container.container.logs(stdout=stdout, stderr=stderr, tail=tail, since=since, stream=True, timestamps=True)
        
        self.__thread = FunctionThreaded(self.__read_stream, name=f"follow logs of container '{self.__docker_container.container.name}'", register_thread=False)
        self.__thread.interrupt_function = self.__logs_stream.close
        self.__thread.start()
    
    def __read_stream(self):
        for log in self.__logs_stream:
            self.__logs.append( (log['time'], log['steam'] == 'stdout', self.__docker_container._get_log_from_log_bytes(log['log'])) )
    
    def reset(self):
        self.__logs.clear()
    
    def get_logs(self, stdout=True, stderr=True, timestamps=False, tail='all', since=None, until=None, wait_before_s=0.1):
        if wait_before_s is not None:
            time.sleep(wait_before_s)
        
        if since is not None and isinstance(since, datetime):
            since = DateTime.datetime_2_str(since)
        if until is not None and isinstance(until, datetime):
            until = DateTime.datetime_2_str(until)
        
        res = []
        for log in self.__logs:
            if log[1] and not stdout or not log[1] and not stderr:
                continue
            if since is not None and log[0] < since:
                continue
            if until is not None and log[0] >= until:
                continue
            
            if timestamps:
                res.append( (log[0], log[2]) )
            else:
                res.append(log[2])
        
        if tail != 'all':
            if not isinstance(tail, int) or tail < 0:
                raise TechnicalException(f"Argument 'tail' can be 'all' or a positive integer")
            res = res[-tail:]
        
        return res




class DockerContainerLogsFormatter(object):
    """Base class for formatters of container logs.
    """
    def __init__(self, include_patterns=None, exclude_patterns=None, include_others_name=default_value):
        """ Constructor
        @param include_patterns: patterns of log fields to include
        @param exclude_patterns: patterns of log fields to exclude
        @param include_others_name: if defined, all fields not included and not excluded are grouped in this field name (used only if include_patterns is defined)
        """
        self.__include_patterns = {p if isinstance(p, re.Pattern) else re.compile(p):pinfo for p, pinfo in include_patterns.items()} if include_patterns is not None else None
        self.__exclude_patterns = [ep if isinstance(ep, re.Pattern) else re.compile(ep) for ep in exclude_patterns] if exclude_patterns is not None else None
        self.__include_others_name = _get_include_others_name(include_others_name)
    
    def _filter_json_log_fields(self, log_dict):
        if self.__include_patterns is None and self.__exclude_patterns is None:
            # No filters are defined
            return log_dict
        
        res = {}
        log = copy.copy(log_dict)
        
        # Manage included fields and remove them from source copy
        if self.__include_patterns is not None:
            for pat, pat_info in self.__include_patterns.items():
                for key in list(log.keys()):
                    if pat.match(key):
                        name = pat_info['replace'] if 'replace' in pat_info else key
                        res[name] = log.pop(key)
        
        # Manage excluded fields and remove them from source copy
        if self.__exclude_patterns is not None:
            for pat in self.__exclude_patterns:
                for key in list(log.keys()):
                    if pat.match(key):
                        del log[key]
        
        # Manage remaining fields of source copy
        if log:
            if self.__include_patterns is not None:
                if self.__include_others_name is not None:
                    # Include remaining fields in surrounding field self.__include_others_name
                    res[self.__include_others_name] = log
            else:
                # Remaining fields are the result
                res = log
        
        return res
    
    def format_logs(self, logs, with_timestamps=False):
        raise NotImplementedError()

class JsonDockerContainerLogsFormatter(DockerContainerLogsFormatter):
    """Formatter of container logs in JSON format.
    """
    def __init__(self, timestamps_field_name=default_value, include_patterns=None, exclude_patterns=None, include_others_name=None):
        """ Constructor
        @param timestamps_field_name: field name of timestamps if they are added to logs
        @param include_patterns: patterns of log fields to include
        @param exclude_patterns: patterns of log fields to exclude
        @param include_others_name: if defined, all fields not included and not excluded are grouped in this field name (used only if include_patterns is defined)
        """
        super().__init__(include_patterns, exclude_patterns, include_others_name)
        self.__timestamps_field_name = _get_timestamps_field_name(timestamps_field_name)
    
    def format_logs(self, logs, with_timestamps=False):
        if with_timestamps and self.__timestamps_field_name is None:
            raise TechnicalException("Timestamps field name must defined to include docker timestamps")
        
        res = []
        
        for log in logs:
            ts = None
            if with_timestamps:
                ts, log = log[0], log[1]
            
            res_log = {}
            
            # Begin by timestamp so that it is the first dict element
            if ts is not None:
                res_log[self.__timestamps_field_name] = ts
            
            # Add log as JSON format
            try:
                log_dict = json.loads(log)
            except:
                # In some circumstances (ex: container is in panic), logs can be in another format even if configured in json format
                log_dict = {'log':log}
            res_log.update(self._filter_json_log_fields(log_dict))
            
            res.append(res_log)
        
        return res




class DockerContainerLogsSaver(object):
    """Base class for savers of docker container logs.
    """
    def __init__(self, timestamps_field_name=default_value, include_patterns=None, exclude_patterns=None, include_others_name=default_value, formatter=None):
        """ Constructor
        @param formatter: logs formatter
        """
        self.__timestamps_field_name = _get_timestamps_field_name(timestamps_field_name)
        self.__include_patterns = {p if isinstance(p, re.Pattern) else re.compile(p):pinfo for p, pinfo in include_patterns.items()} if include_patterns is not None else None
        self.__exclude_patterns = [ep if isinstance(ep, re.Pattern) else re.compile(ep) for ep in exclude_patterns] if exclude_patterns is not None else None
        self.__include_others_name = _get_include_others_name(include_others_name)
        self.__formatter = formatter
        
        if self.__formatter is None:
            self.__formatter = JsonDockerContainerLogsFormatter(timestamps_field_name=timestamps_field_name, include_patterns=include_patterns, exclude_patterns=exclude_patterns, include_others_name=include_others_name)
    
    def get_logs(self, docker_container, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None, wait_before_s=0.1):
        return docker_container.get_logs(stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s, formatter=self.__formatter)
        
    def save_logs(self, docker_container, full_file_name, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None):
        """ Save logs of a container.
        @return: number of logs saved
        """
        raise NotImplementedError()
    
    def _define_field_names_for_logs(self, logs):
        res = []
        
        # Get all field names from logs
        names = list(set().union(*logs))
        
        # Sort by included fields
        if self.__timestamps_field_name and self.__timestamps_field_name in names:
            res.append(self.__timestamps_field_name)
            names.remove(self.__timestamps_field_name)
        if self.__include_patterns:
            for ip, ip_info in self.__include_patterns.items():
                i = 0
                while i < len(names):
                    if ip_info and 'replace' in ip_info:
                        if names[i] == ip_info['replace']:
                            res.append(names.pop(i))
                            continue
                    elif ip.match(names[i]):
                        res.append(names.pop(i))
                        continue
                    
                    # else increment counter
                    i += 1
        
        # Add remaining names in alphabetical order
        if names:
            res.extend(sorted(names))
        
        return res
    
    def _get_log_fields_values(self, field_names, log, missing_field_value=''):
        res = []
        
        for name in field_names:
            if name in log:
                res.append(log[name])
            else:
                res.append(missing_field_value)
        
        return res
    

class JsonDockerContainerLogsSaver(DockerContainerLogsSaver):
    """Saver of docker container logs in a stream JSON file.
    """
    def __init__(self, timestamps_field_name=default_value, include_patterns=None, exclude_patterns=None, include_others_name=default_value, formatter=None, **dumps_kwargs):
        """ Constructor
        @param timestamps_field_name: field name of timestamps if they are added to logs
        @param include_patterns: patterns of log fields to include
        @param exclude_patterns: patterns of log fields to exclude
        @param include_others_name: if defined, all fields not included and not excluded are grouped in this field name (used only if include_patterns is defined)
        @param formatter: formatter to use (if defined, previous fields are omitted)
        @param dumps_kwargs: all additional kwargs are used as json.dumps additional parameters
        """
        super().__init__(timestamps_field_name=timestamps_field_name, include_patterns=include_patterns, exclude_patterns=exclude_patterns, include_others_name=include_others_name, formatter=formatter)
        self.__dumps_kwargs = dumps_kwargs
        
    def save_logs(self, docker_container, full_file_name, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None, wait_before_s=0.1):
        res = None
        
        file_path = f"{full_file_name}.json"
        SessionContext.instance().path_manager.makedirs(file_path)
        
        logs = self.get_logs(docker_container, stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s)
        if logs:
            with StreamJSONFile(file_path, mode='wt') as fout:
                fout.write_elements_json_object_list(logs, **self.__dumps_kwargs)
            res = len(logs)
            logger.info(f"Saved {res} logs for container '{docker_container.container.name}' in file '{file_path}'")
        
        return res

class PrettyTableDockerContainerLogsSaver(DockerContainerLogsSaver):
    """Saver of docker container logs in a pretty table file.
    """
    def __init__(self, timestamps_field_name=default_value, include_patterns=None, exclude_patterns=None, include_others_name=default_value, formatter=None, **prettytable_kwargs):
        """ Constructor
        @param timestamps_field_name: field name of timestamps if they are added to logs
        @param include_patterns: patterns of log fields to include
        @param exclude_patterns: patterns of log fields to exclude
        @param include_others_name: if defined, all fields not included and not excluded are grouped in this field name (used only if include_patterns is defined)
        @param formatter: formatter to use (if defined, previous fields are omitted)
        @param prettytable_kwargs: all additional kwargs are used as PrettyTable constructor parameters
        """
        super().__init__(timestamps_field_name=timestamps_field_name, include_patterns=include_patterns, exclude_patterns=exclude_patterns, include_others_name=include_others_name, formatter=formatter)
        self.__prettytable_kwargs = prettytable_kwargs
        
    def save_logs(self, docker_container, full_file_name, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None, wait_before_s=0.1):
        import prettytable
        from prettytable.prettytable import PrettyTable, TableStyle
        
        res = None
        
        file_path = f"{full_file_name}.txt"
        SessionContext.instance().path_manager.makedirs(file_path)
        
        logs = self.get_logs(docker_container, stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s)
        
        if logs:
            # Build table
            table = PrettyTable(**self.__prettytable_kwargs)
            table.field_names = self._define_field_names_for_logs(logs)
            for log in logs:
                table.add_row(self._get_log_fields_values(table.field_names, log))
            
            # Output table as string
            table.set_style(TableStyle.ORGMODE)
            table.align = 'l'
            table.hrules = prettytable.NONE
            table_str = table.get_string()
            
            with File(file_path, mode='wt') as fout:
                fout.write(table_str)
                fout.write('\n')
            
            res = len(logs)
            logger.info(f"Saved {res} logs for container '{docker_container.container.name}' in file '{file_path}'")
        
        return res

class CsvDockerContainerLogsSaver(DockerContainerLogsSaver):
    """Saver of docker container logs in a CSV file.
    """
    def __init__(self, timestamps_field_name=default_value, include_patterns=None, exclude_patterns=None, include_others_name=default_value, formatter=None, **csv_kwargs):
        """ Constructor
        @param timestamps_field_name: field name of timestamps if they are added to logs
        @param include_patterns: patterns of log fields to include
        @param exclude_patterns: patterns of log fields to exclude
        @param include_others_name: if defined, all fields not included and not excluded are grouped in this field name (used only if include_patterns is defined)
        @param formatter: formatter to use (if defined, previous fields are omitted)
        @param csv_kwargs: all additional kwargs are used as csv.DictWriter additional parameters
        """
        super().__init__(timestamps_field_name=timestamps_field_name, include_patterns=include_patterns, exclude_patterns=exclude_patterns, include_others_name=include_others_name, formatter=formatter)
        self.__csv_kwargs = csv_kwargs
        
    def __get_log_row(self, field_names, log):
        res = []
        
        for name in field_names:
            if name in log:
                res.append(log[name])
            else:
                res.append('')
        
        return res
    
    def save_logs(self, docker_container, full_file_name, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None, wait_before_s=0.1):
        import csv
        
        res = None
        
        file_path = f"{full_file_name}.csv"
        SessionContext.instance().path_manager.makedirs(file_path)
        
        logs = self.get_logs(docker_container, stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s)
        
        if logs:
            field_names = self._define_field_names_for_logs(logs)
            
            with File(file_path, mode='wt') as fout:
                dw = csv.DictWriter(fout, fieldnames=field_names, **self.__csv_kwargs)
                
                dw.writeheader()
                for log in logs:
                    dw.writerow(log)
            
            res = len(logs)
            logger.info(f"Saved {res} logs for container '{docker_container.container.name}' in file '{file_path}'")
        
        return res



class DockerContainersLogsSaver():
    """Saver of logs for a group of containers
    """
    def __init__(self, docker_client, container_logs_saver=default_value, include_patterns=None, exclude_patterns=None):
        self.__docker_client = docker_client
        self.__container_logs_saver = container_logs_saver if container_logs_saver is not default_value else JsonDockerContainerLogsSaver()
        self.__include_patterns = [ip if isinstance(ip, re.Pattern) else re.compile(ip) for ip in include_patterns] if include_patterns is not None else None
        self.__exclude_patterns = [ep if isinstance(ep, re.Pattern) else re.compile(ep) for ep in exclude_patterns] if exclude_patterns is not None else None
        
    def save_containers_logs(self, destination_path, stdout=True, stderr=True, timestamps=True, tail='all', since=None, until=None, wait_before_s=0.1):
        """ Save containers logs
        @return: number of containers for which logs have been saved
        """
        res = 0
        has_waited = False  # Manage to wait only on first container
        
        names = self.__docker_client.get_container_names(in_list=True, all_=True, sparse=True, include_patterns=self.__include_patterns, exclude_patterns=self.__exclude_patterns)
        for name in names:
            container = self.__docker_client.get_container(name, all_=True, reset_if_removed=False)
            full_file_name = os.path.join(destination_path, name)
            nb_logs = self.__container_logs_saver.save_logs(container, full_file_name, stdout=stdout, stderr=stderr, timestamps=timestamps, tail=tail, since=since, until=until, wait_before_s=wait_before_s if not has_waited else None)
            has_waited = True
            
            if nb_logs is not None:
                res += 1
        
        return res





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import abc
import logging
from typing import Generator, overload
from holado_core.common.exceptions.functional_exception import FunctionalException

logger = logging.getLogger(__name__)


class BaseGenerator(Generator):
    '''
    Base class for generators.
    It extends the possibilities of python generators with batch generation.
    Note: current implementation returns a batch as a list, so be careful on batch size.
    '''
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name=None, batch_size=None):
        super().__init__()
        self.__name = name if name is not None else "generator"
        self.__batch_size = batch_size
        self.__is_in_next_batch = False
        
    @property
    def name(self):
        return self.__name
    
    @property
    def batch_size(self):
        return self.__batch_size
    
    @batch_size.setter
    def batch_size(self, size):
        self.__batch_size = size
    
    @property
    def _is_in_next_batch(self):
        return self.__is_in_next_batch
    
    @overload
    # @abstractmethod
    def send(self, value):
        return Generator.send(self, value)
    
    @overload
    # @abstractmethod
    def throw(self, typ, val=None, tb=None):
        return Generator.throw(self, typ, val=val, tb=tb)
    
    def next_batch(self, size=None, raise_if_incomplete=False, raise_if_empty=True):
        """ Return a batch as a liste of defined size.
        If size parameter is not defined, batch size is supposed to be defined at generator instantiation or with property batch_size.
        If raise_if_incomplete is True, and last batch can not be of batch size, a StopIteration is thrown.
        """
        if size is None:
            if self.batch_size is None:
                raise FunctionalException(f"Undefined batch size. It can be defined at generator instantiation, with property batch_size.")
            size = self.batch_size
        
        self.__is_in_next_batch = True
        try:
            res = []
            for _ in range(size):
                try:
                    element = next(self)
                except StopIteration as exc:
                    if raise_if_empty and len(res) == 0 or raise_if_incomplete:
                        raise exc
                    else:
                        break
                else:
                    # if element is not None:
                        res.append(element)
        finally:
            self.__is_in_next_batch = False
        
        return res






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from timeit import default_timer as timer
from holado_core.common.finders.finder import Finder
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.element_exception import NoSuchElementException
from holado.holado_config import Config
import logging
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class OrFinder(Finder):
    """ Finder that search using multiple finder.
    All referenced finders are executed, and all results are merged.
    """
    
    def __init__(self, description=None):
        super().__init__(description)
        self.__finders = []
    
    @property
    def finders(self):
        return self.__finders
    
    def add_finder(self, finder):
        """
        Append given finder in or succession.
        @param finder A finder
        """
        self.__finders.append(finder)

    def add_finders(self, other_finder):
        """
        @param other_finder Other OrFinder to merge with
        """
        for finder in other_finder.finders:
            self.add_finder(finder)

    def _find_all(self, find_context, find_parameters):
        context = self.update_context(find_context)
        parameters = self.update_parameters(find_parameters)
        
        res = []
        exc_previous = None
        sub_find_parameters = parameters.get_next_level()
        
        # Prepare analyze time spent
        if parameters.analyze_time_spent:
            start = Tools.timer_s()

        for index, finder in enumerate(self.__finders):
            try:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{self._get_indent_string_level(parameters)}[OrFinder({self.element_description})[{index}]:{finder.element_description}] begin")
                beg = Tools.timer_s()
                
                try:
                    self._add_candidates(res, finder.find_all(context, sub_find_parameters))
                except NoSuchElementException as exc:
                    if exc_previous is None:
                        exc_previous = exc
                
                end = Tools.timer_s()
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"{self._get_indent_string_level(parameters)}[OrFinder({self.element_description})[{index}]:{finder.element_description}] end ({end-beg} s)")
            except TechnicalException as exc:
                raise TechnicalException(f"[OrFinder({self.element_description})[{index}]:{finder.element_description}] {exc.message}") from exc
        
        # Analyze time spent
        if parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"OrFinder[{self.element_description}].findAll end (took {duration} s)     ->  {len(res)} elements")

        if len(res) == 0 and exc_previous is not None:
            Tools.raise_same_exception_type(exc_previous, f"Unable to find {self.element_description} in {find_context.get_find_type_container_description_prefix()}{find_context.get_input_complete_description_and_details()}")
        return res
    
    def is_valid_input(self, container, find_context, find_parameters):
        return self._is_valid_input_in(self.__finders, container, find_context, find_parameters)
    
    def is_valid_output(self, element, find_context, find_parameters):
        return self._is_valid_output_in(self.__finders, element, find_context, find_parameters)
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
import time
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_multitask.multiprocessing.process import InterruptableProcess
from holado_multitask.multiprocessing.function_process import FunctionProcess

logger = logging.getLogger(__name__)


class PeriodicFunctionProcess(InterruptableProcess):
    '''
    Execute periodically in a process given function with given arguments.
    
    Implementation uses a precise timer, in order to globally satisfy the desired periodicity (with a precision of around a nanosecond, depending to system latency).
    But implementation is quite simple, function is executed around each timer ticks : it is based on waits after function execution of the time needed to satisfy periodicity.
    If function execution is greater than periodicity, a warning is logged, and function is run immediately without waiting next timer tick.
    
    For fast periodicity, it is recommended to disable logs by setting property 'do_log' to False (and also 'do_log_warning' if needed).
    '''
    
    def __init__(self, period_seconds, target, args=None, kwargs=None, name=None, default_wait_timeout=None, register_process=True, delay_before_run_sec=None, nb_times=None):
        super().__init__(name if name is not None else f"PeriodicFunctionProcess({repr(target)})", default_wait_timeout=default_wait_timeout, register_process=register_process, delay_before_run_sec=delay_before_run_sec, delay_between_run_sec=None)

        self.__period_seconds = period_seconds
        if self.__period_seconds < 1e-8:
            raise TechnicalException(f"Periodicity must be at least 10ns")
        self.__nb_times = nb_times
        self.__function_process = FunctionProcess(target, args, kwargs)
        self.__do_log = True
        self.__do_log_warning = True
        
        if isinstance(self.__period_seconds, int) or self.__period_seconds % 1 < 1e-9 or self.__period_seconds % 1 > 1 - 1e-9:
            self.__timer = Tools.timer_s
            self.__timer_period = self.__period_seconds
            self.__timer_unity = "s"
            self.__timer_s_value = 1
        else:
            self.__timer = Tools.timer_ns
            self.__timer_period = self.__period_seconds * 1e9
            self.__timer_unity = "ns"
            self.__timer_s_value = 1e9

    @property
    def function_process(self):
        return self.__function_process

    @property
    def do_log(self):
        return self.__do_log

    @do_log.setter
    def do_log(self, do_log):
        self.__do_log = do_log

    @property
    def do_log_warning(self):
        return self.__do_log_warning

    @do_log_warning.setter
    def do_log_warning(self, do_log):
        self.__do_log_warning = do_log

    def run(self):
        # Start running by superclass run
        super().run()
        
        logger.info(f"[{self.name}] Process managing periodic function.")
        
        self.wait_before_run()
        
        start = self.__timer()
        count = 0
        while not self.is_interrupted:
            # Process function
            if self.do_log:
                logger.info(f"[{self.name}] Processing periodic iteration {count+1}...")
            beg = self.__timer()
            try:
                self.__function_process.run()
            except Exception as exc:
                msg = "{} Exception catched while processing function:\n    [{}]".format(self.name, repr(exc))
                logger.exception(msg)
            finally:
                count += 1
                iter_dur = self.__timer() - beg
                
            # Define next iteration timer
            next_timer = start + count * self.__timer_period
            wait_duration = next_timer - self.__timer()
            skip_wait = wait_duration < 0
            if skip_wait:
                if self.do_log_warning:
                    logger.warning(f"[{self.name}] {count}'th iteration finished too late (of {-wait_duration} {self.__timer_unity}) to process next iteration on time (duration: {iter_dur} {self.__timer_unity} ; periodicity: {self.__timer_period} {self.__timer_unity}). Processing immediatelly next iteration")
                
                while wait_duration + self.__timer_period < 0:
                    if self.do_log_warning:
                        logger.warning(f"[{self.name}] Failed to process {count+1}'th iteration, its activity period is passed")
                    count += 1
                    next_timer = start + count * self.__timer_period
                    wait_duration = next_timer - self.__timer()
            
            # Stop if nb_times is reached
            if self.__nb_times is not None:
                if count >= self.__nb_times:
                    break
            
            # Wait periodicity
            if not skip_wait:
                if self.do_log:
                    logger.info(f"[{self.name}] Waiting {wait_duration} {self.__timer_unity} to process next periodic function...")
                while not self.is_interrupted:
                    if wait_duration > 0.1 * self.__timer_s_value:
                        time.sleep(0.1)
                        wait_duration = next_timer - self.__timer()
                    else:
                        if wait_duration > 1e-9 * self.__timer_s_value:
                            time.sleep(wait_duration / self.__timer_s_value)
                        break
            
        logger.info(f"[{self.name}] Process of periodic function has finished.")
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table_row import TableRow
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.converters.converter import Converter
from holado_python.standard_library.typing import Typing
import logging
from holado_json.ipc.json_converter import JsonConverter
from holado_core.common.tables.table_manager import TableManager
from holado_core.common.exceptions.functional_exception import FunctionalException
import re
from holado_core.common.tools.tools import Tools
from holado_data.data.generator.base import BaseGenerator


logger = logging.getLogger(__name__)



class TableConverter(object):
    
    @classmethod
    def convert_table_with_header_to_dict(cls, table):
        res_list = cls.convert_table_with_header_to_dict_list(table, as_generator=False)
        if len(res_list) == 0:
            return None
        elif len(res_list) == 1:
            return res_list[0]
        else:
            raise TechnicalException(f"Failed to convert table to dict, since table has more than one row.")
    
    @classmethod
    def convert_table_with_header_to_dict_list(cls, table, as_generator=False):
        if as_generator:
            class TableWithHeader2DictGenerator(BaseGenerator):
                def __init__(self, table):
                    super().__init__(name="table with header to dict generator")
                    self.__table = table
                    self.__index_by_name = table.get_column_indexes_by_string_content()
                    self.__table_rows_iter = iter(self.__table.rows)
                
                def __next__(self):
                    row = next(self.__table_rows_iter)
                    return {name: row.get_cell(index).content for name, index in self.__index_by_name.items()}
            
            return TableWithHeader2DictGenerator(table)
        else:
            # index_by_name = table.get_column_indexes_by_string_content()
            #
            # res = []
            # for row in table.rows:
            #     new_dict = {name: row.get_cell(index).content for name, index in index_by_name.items()}
            #     res.append(new_dict)
            #
            # return res
            gen = cls.convert_table_with_header_to_dict_list(table, as_generator=True)
            return [e for e in gen]
    
    @classmethod
    def convert_table_2_list_of_tuples(cls, table, as_generator=False):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Converting table to a list of tuples...")
        if as_generator:
            class Table2TupleGenerator(BaseGenerator):
                def __init__(self, table):
                    super().__init__(name="table with header to dict generator")
                    self.__table = table
                    self.__table_rows_iter = iter(self.__table.rows)
                
                def __next__(self):
                    row = next(self.__table_rows_iter)
                    return tuple((cell.content for cell in row.cells))
            
            return Table2TupleGenerator(table)
        else:
            # return [tuple((cell.content for cell in row.cells)) for row in table.rows]
            gen = cls.convert_table_2_list_of_tuples(table, as_generator=True)
            return [e for e in gen]
    
    @classmethod
    def convert_dict_list_to_table_with_header(cls, dict_list):
        """
        Builds a table with header from a list of dictionaries.
        """
        header = set()
        for index, new_dict in enumerate(dict_list):
            if not Converter.is_dict(new_dict):
                raise TechnicalException(f"The list element of index {index} is not a dict (obtained type: {Typing.get_object_class_fullname(new_dict)})")
            header = header.union(new_dict.keys())
        header = sorted(header)

        res = TableWithHeader()
        res.header = TableRow(cells_content=header)
        
        for new_dict in dict_list:
            res.add_row(contents_by_colname=new_dict)
        
        return res
    
    @classmethod
    def convert_name_value_table_2_dict(cls, table):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Name/Value table to dict (table = {table})")

        res = {}
        if table is not None:
            # Verify table structure
            TableManager.verify_table_is_name_value_table(table)
            
            for row in table.rows:
                if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace("Converting row (%s)", row)
                
                name = row.get_cell(0).content
                value = row.get_cell(1).content
                
                if name in res:
                    raise FunctionalException("Name '{}' appears several times in table".format(name))
                else:
                    res[name] = value
                
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting Name/Value table to dict (table = {table}) => {res}")
        return res

    @classmethod
    def convert_name_value_table_2_json_object(cls, table, converter=None):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Converting Name/Value table to json object (table = {table})")
        if converter is None:
            converter = JsonConverter()
        
        # Verify table structure
        TableManager.verify_table_is_name_value_table(table)
        
        res = {}
        for row in table.rows:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Converting row (%s)", row)
            
            name = row.get_cell(0).content
            value = row.get_cell(1).content
            value = converter.to_json(value)
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Adding for name '{name}' json value [{value}] (type: {Typing.get_object_class_fullname(value)})")
            cls._fill_json_dict(res, name, value)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Converting Name/Value table to json object (table = {table}) => {res}")
        return res
 
    @classmethod
    def extract_table_with_header_from_table(cls, table, row_first_index, row_last_index, col_first_index, col_last_index):
        res = TableWithHeader()
        
        header_content = [c.content for c in table.rows[row_first_index].cells[col_first_index:col_last_index+1]]
        res.header = TableRow(cells_content=header_content)
        
        for i in range(row_first_index+1, row_last_index+1):
            row_content = [c.content for c in table.rows[i].cells[col_first_index:col_last_index+1]]
            res.add_row(cells_content=row_content)
        
        return res
        
 
    @classmethod
    def _fill_json_dict(cls, json_res, name, value):
        if not isinstance(json_res, dict):
            raise TechnicalException(f"Type '{Typing.get_object_class_fullname(json_res)}' is not a dict")
        
        names = name.split(".", maxsplit=1)
        name0 = names[0]
        name1 = names[1] if len(names) > 1 else None
        
        # manage lists
        m = re.match(r"^([^[]+)\[(.*)\]$", name0)
        if m:
            li_name = m.group(1)
            index = int(m.group(2))

            if li_name not in json_res:
                json_res[li_name] = []
            cls._fill_json_list(json_res[li_name], index, name1, value)
            return
        
        if name1 is None:
            cls._fill_json_dict_entry(json_res, name0, value)
        else:
            if name0 not in json_res:
                json_res[name0] = {}
            cls._fill_json_dict(json_res[name0], name1, value)
 
    @classmethod
    def _fill_json_dict_entry(cls, json_res, name, value):
        if name in json_res:
            raise FunctionalException(f"Name '{name}' is already set")
        else:
            json_res[name] = value
 
    @classmethod
    def _fill_json_list(cls, json_res, index, name, value):
        if not isinstance(json_res, list):
            raise TechnicalException(f"Type '{Typing.get_object_class_fullname(json_res)}' is not a list")
        
        # Prepare list with missing elements
        if index + 1 > len(json_res):
            for _ in range(len(json_res), index + 1):
                json_res.append(None)
        
        if name is None:
            if json_res[index] is None:
                json_res[index] = value
            else:        
                raise FunctionalException("List has already an element at index {}".format(index))
        else:
            if json_res[index] is None:
                json_res[index] = {}
            cls._fill_json_dict(json_res[index], name, value)
            
        
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_system.system.command.exceptions import CommandError

logger = logging.getLogger(__name__)


class CommandResult(object):
    """
    Result of a command
    """

    def __init__(self, command, output=None, error=None, previous_command_result=None):
        self.__command = command
        self.__output = output
        self.__error = error

        if previous_command_result is not None:
            # Verify command is the same
            if previous_command_result.command != command:
                raise TechnicalException(f"Commands are not the same:\n    current:  [{command}]\n    previous: [{previous_command_result.command}]")
            
            if output is not None and previous_command_result.output is not None:
                self.__output = previous_command_result.output + output
            elif output is not None:
                self.__output = output
            else:
                self.__output = previous_command_result.output
            
            if error is not None and previous_command_result.error is not None:
                self.__error = previous_command_result.error + error
            elif error is not None:
                self.__error = error
            else:
                self.__error = previous_command_result.error
    
    @property
    def command(self):
        """
        @return Command
        """
        return self.__command
    
    @property
    def output(self):
        """
        @return Command output
        """
        return self.__output
 
    @output.setter
    def output(self, output):
        """
        @param output Command output
        """
        self.__output = output
    
    @property
    def error(self):
        """
        @return Command error
        """
        return self.__error
 
    @error.setter
    def error(self, error):
        """
        @param error Command error
        """
        self.__error = error

    @property
    def has_error(self):
        """
        @return True if error exists
        """
        return (self.error is not None and len(self.error) > 0)
    
    @property
    def has_output(self):
        """
        @return True if output exists
        """
        return (self.output is not None and len(self.output) > 0)
    
    def log_if_error(self):
        """
        Log an error if command resulted with an error.
        """
        if self.has_error:
            logger.error(f"Error during command '{self.command}': {self.error}")
    
    def raise_if_error(self):
        """
        Raise a CommandError if command resulted with an error.
        """
        if self.has_error:
            raise CommandError(self.command, self.error)
    
    def __eq__(self, obj:object)->bool:
        if obj is None or not isinstance(obj, type(self)):
            return False
        
        res = self.command is None and obj.command is None \
                or self.command is not None and self.command == obj.command
        
        res &= self.output is None and obj.output is None \
                or self.output is not None and self.output == obj.output
        
        res &= self.error is None and obj.error is None \
                or self.error is not None and self.error == obj.error
        
        return res
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.actors.actions import FindAction
from holado_ui.ui.gui.actors.gui_actor import GUIActor
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_redo import SeleniumRedo
from holado_multitask.multithreading.reflection.sys import get_current_function_name
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class SeleniumActor(GUIActor):
    """ Selenium actor.
    """
    
    def __init__(self, module_name):
        super().__init__(module_name)
        
        self.__click_with_javascript = False
    
    @property
    def click_with_javascript(self):
        return self.__click_with_javascript
    
    @click_with_javascript.setter
    def click_with_javascript(self, with_javascript):
        """
        @param with_javascript True if click must be processed with Javascript rather than with Selenium
        """
        self.__click_with_javascript = with_javascript
    
    def click_on_element(self, element, scroll_by_x=0, scroll_by_y=0):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Clicking on element with scrollBy ({scroll_by_x}, {scroll_by_y})...")
        
        # Scroll to element, so that click action is visible on screenshot after an error
        if scroll_by_x != 0 or scroll_by_y != 0:
            self.internal_api.scroll_into_view(element, 0)
            self.internal_api.scroll_by(scroll_by_x, scroll_by_y, 100)
        else:
            self.internal_api.scroll_into_view(element)
        
        # Move mouse over
        if self.do_mouse_over_before_click:
            self.internal_api.moveMouseOver(element)
        
        # Click on element
        if self.click_with_javascript:
            self.internal_api.click_on_element_with_javascript(element)
        else:
            self.internal_api.click_on_element_with_selenium(element)
        
        self.internal_api.wait_until_window_is_loaded()
    
    def select_value(self, element, value):
        """
        Select a value in element
        @param element Element
        @param value Text
        @throws FunctionalException
        """
        finder_info = element.find_info.finder.info
        self._call_action_from_module(finder_info, get_current_function_name(), element, value)
    
    def write_to_element(self, element, value):
        if self.do_write_to_element_key_by_key:
            self.internal_api.write_to_element_key_by_key(element, value)
        else:
            self.internal_api.write_to_element_all_keys(element, value)
    
    def act_find_and_click_on(self, finder, find_type):
        def func(container, candidates, find_context, find_parameters):
            new_parameters = find_parameters.with_redo(False)
            
            class ActRedo(SeleniumRedo):
                def __init__(self_redo):  # @NoSelf
                    super().__init__(f"find and click on {finder.get_element_description()}", self.driver)
                    
                def _process(self_redo):  # @NoSelf
                    element = finder.find(find_type, container, candidates, find_context, new_parameters)
                    self.click_on_element(element, 0, self_redo.scroll_by_y)
                    return None
            
            redo = ActRedo()
            redo.with_iteration_on_scroll_by_y()
            redo.ignore_all(find_parameters.redo_ignored_exceptions)
            redo.execute()
            return None
        
        return FindAction(f"find {finder.get_element_description()} and click on", self.find_builder, func)
    
    def _get_redo(self, action, *args, **kwargs):
        """
        Redo an action
        @param action Action
        """
        class ActionRedo(SeleniumRedo):
            def __init__(self):
                super().__init__(f"redo {action.name}")
                
            def _process(self):
                return action.execute(*args, **kwargs)
            
            def _get_waited_description(self):
                return action.name
            
        return ActionRedo()
    
    
    
    
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import os.path
import sys
import logging


# Add testing solution sources paths
here = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, here)

# Add HolAdo source paths (needed when using a clone of HolAdo project)
from initialize_holado import insert_holado_source_paths  # @UnresolvedImport
insert_holado_source_paths()


# Configure HolAdo
import holado
is_in_steps_catalog = holado.is_in_steps_catalog()
if is_in_steps_catalog:
    holado.initialize(TSessionContext="test_holado.test_holado_session_context.TestHoladoSessionContext", logging_config_file_path=os.path.join(here, 'logging.conf'), 
                      log_level=logging.INFO, log_on_console=True, 
                      log_in_file=False, session_kwargs={"with_session_path":False})
else:
    # Initialize HolAdo:
    #    - log_level is set to INFO for initialization phase, it will be overwrite by level in logging config file
    #    - log_on_console is True for initialization phase, it will be set to False when root log file will be defined
    #    - logging config file
    holado.initialize(TSessionContext="test_holado.test_holado_session_context.TestHoladoSessionContext", logging_config_file_path=os.path.join(here, 'logging.conf'), 
                      log_level=logging.INFO, log_on_console=True,
                      config_kwargs={'application_group':'test_runner'})



# Import generic environment methods
from holado_test.behave.behave_environment import *  # @UnusedWildImport

# Define project specific environment methods

if not is_in_steps_catalog:
    # Wait test daemons are healthy
    from holado_core.common.tools.converters.converter import Converter
    do_wait_test_server = Converter.to_boolean(os.getenv("HOLADO_WAIT_TEST_SERVER", True))
    if do_wait_test_server:
        SessionContext.instance().test_server_client.wait_is_healthy()


